class TEXTWORLD_REMOTE_ENV:
    PING = "TEXTWORLD_REMOTE_ENV.PING"
    PONG = "TEXTWORLD_REMOTE_ENV.PONG"
    ENV_CREATE = "TEXTWORLD_REMOTE_ENV.ENV_CREATE"
    ENV_CREATE_RESPONSE = "TEXTWORLD_REMOTE_ENV.ENV_CREATE_RESPONSE"
    ENV_RESET = "TEXTWORLD_REMOTE_ENV.ENV_RESET"
    ENV_RESET_RESPONSE = "TEXTWORLD_REMOTE_ENV.ENV_RESET_RESPONSE"
    ENV_STEP = "TEXTWORLD_REMOTE_ENV.ENV_STEP"
    ENV_STEP_RESPONSE = "TEXTWORLD_REMOTE_ENV.ENV_STEP_RESPONSE"
    ENV_SUBMIT = "TEXTWORLD_REMOTE_ENV.ENV_SUBMIT"
    ENV_SUBMIT_RESPONSE = "TEXTWORLD_REMOTE_ENV.ENV_SUBMIT_RESPONSE"
    ERROR = "TEXTWORLD_REMOTE_ENV.ERROR"
